from elimity_insights_client.api._encode_expression import (
    encode_boolean_expression,
    encode_date_expression,
    encode_date_time_expression,
    encode_number_expression,
    encode_string_expression,
    encode_time_expression,
)
from elimity_insights_client.api.query import (
    AnyExpression,
    BooleanAnyExpression,
    DateAnyExpression,
    DateTimeAnyExpression,
    DirectedGroupOrdering,
    Direction,
    DirectLinkGroupByQuery,
    DirectLinkQuery,
    Grouping,
    GroupOrdering,
    GroupOrderingType,
    LinkGroupByQuery,
    NumberAnyExpression,
    Ordering,
    Query,
    StringAnyExpression,
    TimeAnyExpression,
    UnspecifiedGroupOrdering,
)


def encode_query(query: Query) -> object:
    condition = encode_boolean_expression(query.condition)
    direct_link_group_by_queries = map(
        _encode_direct_link_group_by_query, query.direct_link_group_by_queries
    )
    direct_link_queries = map(_encode_direct_link_query, query.direct_link_queries)
    include = map(_encode_any_expression, query.include)
    link_group_by_queries = map(
        _encode_link_group_by_query, query.link_group_by_queries
    )
    link_queries = map(encode_query, query.link_queries)
    order_by = map(_encode_ordering, query.order_by)
    return {
        "alias": query.alias,
        "condition": condition,
        "directLinkGroupByQueries": direct_link_group_by_queries,
        "directLinkQueries": direct_link_queries,
        "entityType": query.entity_type,
        "include": include,
        "limit": query.limit,
        "linkGroupByQueries": link_group_by_queries,
        "linkQueries": link_queries,
        "offset": query.offset,
        "orderBy": order_by,
        "sourceId": query.source_id,
    }


def _encode_any_expression(expression: AnyExpression) -> object:
    if isinstance(expression, BooleanAnyExpression):
        expr = encode_boolean_expression(expression.expr)
        return {"expr": expr, "type": "boolean"}

    if isinstance(expression, DateAnyExpression):
        expr = encode_date_expression(expression.expr)
        return {"expr": expr, "type": "date"}

    if isinstance(expression, DateTimeAnyExpression):
        expr = encode_date_time_expression(expression.expr)
        return {"expr": expr, "type": "dateTime"}

    if isinstance(expression, NumberAnyExpression):
        expr = encode_number_expression(expression.expr)
        return {"expr": expr, "type": "number"}

    if isinstance(expression, StringAnyExpression):
        expr = encode_string_expression(expression.expr)
        return {"expr": expr, "type": "string"}

    if isinstance(expression, TimeAnyExpression):
        expr = encode_time_expression(expression.expr)
        return {"expr": expr, "type": "time"}


def _encode_direction(direction: Direction) -> object:
    if direction is Direction.ASC:
        return "asc"

    if direction is Direction.DESC:
        return "desc"


def _encode_direct_link_group_by_query(query: DirectLinkGroupByQuery) -> object:
    condition = encode_boolean_expression(query.condition)
    group_by = map(_encode_grouping, query.group_by)
    return {
        "alias": query.alias,
        "condition": condition,
        "entityType": query.entity_type,
        "groupBy": group_by,
        "linkAlias": query.link_alias,
        "sourceId": query.source_id,
    }


def _encode_direct_link_query(query: DirectLinkQuery) -> object:
    condition = encode_boolean_expression(query.condition)
    direct_link_group_by_queries = map(
        _encode_direct_link_group_by_query, query.direct_link_group_by_queries
    )
    direct_link_queries = map(_encode_direct_link_query, query.direct_link_queries)
    include = map(_encode_any_expression, query.include)
    link_group_by_queries = map(
        _encode_link_group_by_query, query.link_group_by_queries
    )
    link_queries = map(encode_query, query.link_queries)
    order_by = map(_encode_ordering, query.order_by)
    return {
        "alias": query.alias,
        "condition": condition,
        "directLinkGroupByQueries": direct_link_group_by_queries,
        "directLinkQueries": direct_link_queries,
        "entityType": query.entity_type,
        "include": include,
        "limit": query.limit,
        "linkAlias": query.link_alias,
        "linkGroupByQueries": link_group_by_queries,
        "linkQueries": link_queries,
        "offset": query.offset,
        "orderBy": order_by,
        "sourceId": query.source_id,
    }


def _encode_grouping(grouping: Grouping) -> object:
    key = _encode_any_expression(grouping.key)
    ordering = _encode_group_ordering(grouping.ordering)
    return {"key": key, "ordering": ordering}


def _encode_group_ordering(ordering: GroupOrdering) -> object:
    if isinstance(ordering, DirectedGroupOrdering):
        direction = _encode_direction(ordering.direction)
        type = _encode_group_ordering_type(ordering.type)
        return {"direction": direction, "type": type}

    if isinstance(ordering, UnspecifiedGroupOrdering):
        return {"type": "none"}


def _encode_group_ordering_type(type: GroupOrderingType) -> object:
    if type is GroupOrderingType.COUNT:
        return "count"

    if type is GroupOrderingType.LABEL:
        return "label"


def _encode_link_group_by_query(query: LinkGroupByQuery) -> object:
    condition = encode_boolean_expression(query.condition)
    group_by = map(_encode_grouping, query.group_by)
    return {
        "alias": query.alias,
        "condition": condition,
        "entityType": query.entity_type,
        "groupBy": group_by,
        "sourceId": query.source_id,
    }


def _encode_ordering(ordering: Ordering) -> object:
    any_expression = _encode_any_expression(ordering.any_expression)
    direction = _encode_direction(ordering.direction)
    return {
        "anyExpression": any_expression,
        "direction": direction,
    }



from typing import List

from elimity_insights_client._domain_graph_schema import (
    AttributeType,
    DomainGraphSchema,
)


def attribute_types(entity_type: str, schema: DomainGraphSchema) -> List[AttributeType]:
    return [
        type
        for type in schema.attribute_types
        if not type.archived and type.entity_type == entity_type
    ]


def link_entity_types(entity_type: str, schema: DomainGraphSchema) -> List[str]:
    return [type.id for type in schema.entity_types if type.id != entity_type]



from typing import List

from more_itertools import interleave

from elimity_insights_client._domain_graph_schema import (
    AttributeType,
    DomainGraphSchema,
    Type,
)
from elimity_insights_client._util import map_list
from elimity_insights_client.api.entities._schema import (
    attribute_types as schema_attribute_types,
)
from elimity_insights_client.api.entities._schema import (
    link_entity_types as schema_link_entity_types,
)
from elimity_insights_client.api.expression import (
    AssignedBooleanExpression,
    AttributeBooleanExpression,
    AttributeDateExpression,
    AttributeDateTimeExpression,
    AttributeNumberExpression,
    AttributeStringExpression,
    AttributeTimeExpression,
    BooleanExpression,
    DateExpression,
    DateTimeExpression,
    LiteralBooleanExpression,
    NumberExpression,
    StringExpression,
    TimeExpression,
)
from elimity_insights_client.api.query import (
    AnyExpression,
    BooleanAnyExpression,
    DateAnyExpression,
    DateTimeAnyExpression,
    DirectLinkGroupByQuery,
    DirectLinkQuery,
    LinkGroupByQuery,
    NumberAnyExpression,
    Ordering,
    Query,
    StringAnyExpression,
    TimeAnyExpression,
)


def query(
    entity_type: str,
    schema: DomainGraphSchema,
    source_id: int,
) -> Query:
    def make_link_query(entity_type: str) -> Query:
        link_queries: List[Query] = []
        return _query(entity_type, link_queries, schema, source_id)

    link_entity_types = schema_link_entity_types(entity_type, schema)
    link_queries = map_list(make_link_query, link_entity_types)
    return _query(entity_type, link_queries, schema, source_id)


def _assigned_inclusion(type: AttributeType) -> AnyExpression:
    expr = AssignedBooleanExpression(type.id, "")
    return BooleanAnyExpression(expr)


def _attribute_inclusion(type: AttributeType) -> AnyExpression:
    id = type.id
    typ = type.type
    if typ is Type.BOOLEAN:
        boolean_expr: BooleanExpression = AttributeBooleanExpression(id, "")
        return BooleanAnyExpression(boolean_expr)
    if typ is Type.DATE:
        date_expr: DateExpression = AttributeDateExpression(id, "")
        return DateAnyExpression(date_expr)
    if typ is Type.DATE_TIME:
        date_time_expr: DateTimeExpression = AttributeDateTimeExpression(id, "")
        return DateTimeAnyExpression(date_time_expr)
    if typ is Type.NUMBER:
        number_expr: NumberExpression = AttributeNumberExpression(id, "")
        return NumberAnyExpression(number_expr)
    if typ is Type.STRING:
        string_expr: StringExpression = AttributeStringExpression(id, "")
        return StringAnyExpression(string_expr)
    time_expr: TimeExpression = AttributeTimeExpression(id, "")
    return TimeAnyExpression(time_expr)


def _inclusions(entity_type: str, schema: DomainGraphSchema) -> List[AnyExpression]:
    attribute_types = schema_attribute_types(entity_type, schema)
    assigned_iter = map(_assigned_inclusion, attribute_types)
    attribute_iter = map(_attribute_inclusion, attribute_types)
    inclusion_iter = interleave(assigned_iter, attribute_iter)
    return list(inclusion_iter)


def _query(
    entity_type: str,
    link_queries: List[Query],
    schema: DomainGraphSchema,
    source_id: int,
) -> Query:
    condition = LiteralBooleanExpression(True)
    direct_link_group_by_queries: List[DirectLinkGroupByQuery] = []
    direct_link_queries: List[DirectLinkQuery] = []
    inclusions = _inclusions(entity_type, schema)
    link_group_by_queries: List[LinkGroupByQuery] = []
    orderings: List[Ordering] = []
    return Query(
        "",
        condition,
        direct_link_group_by_queries,
        direct_link_queries,
        entity_type,
        inclusions,
        999999,
        link_group_by_queries,
        link_queries,
        0,
        orderings,
        source_id,
    )



from datetime import datetime
from typing import Callable, List, TypeVar

from dateutil.tz import tzlocal
from dateutil.utils import default_tzinfo
from simplejson import JSONEncoder

encoder = JSONEncoder(iterable_as_array=True)
local_timezone = tzlocal()

_T = TypeVar("_T")
_U = TypeVar("_U")


def encode_datetime(datetime: datetime) -> object:
    """Encode the given datetime to a JSON value."""
    dat = default_tzinfo(datetime, local_timezone)
    return dat.isoformat()


def map_list(callable: Callable[["_T"], "_U"], iterable: List["_T"]) -> List["_U"]:
    """Apply the given function to each item in the given list, and construct a new list from the results."""
    iterator = map(callable, iterable)
    return list(iterator)



from typing import Optional, TypedDict

from dateutil.parser import isoparse
from typing_extensions import NotRequired

from elimity_insights_client._decode_domain_graph_schema import (
    DomainGraphSchemaDict,
    decode_domain_graph_schema,
)
from elimity_insights_client.api.source import (
    AbsentLastReloadTimestamp,
    LastReloadTimestamp,
    PresentLastReloadTimestamp,
    Source,
)


class SourceDict(TypedDict):
    """JSON value representing a source."""

    archived: bool
    domainGraphSchema: DomainGraphSchemaDict
    id: int
    lastReloadTimestamp: NotRequired[str]
    name: str


def decode_source(dict: SourceDict) -> Source:
    """Decode the given JSON value to a source."""
    archived = dict["archived"]
    domain_graph_schema_dict = dict["domainGraphSchema"]
    domain_graph_schema = decode_domain_graph_schema(domain_graph_schema_dict)
    id = dict["id"]
    last_reload_timestamp_json = dict.get("lastReloadTimestamp")
    last_reload_timestamp = _decode_last_reload_timestamp(last_reload_timestamp_json)
    name = dict["name"]
    return Source(archived, domain_graph_schema, id, last_reload_timestamp, name)


def _decode_last_reload_timestamp(value: Optional[str]) -> LastReloadTimestamp:
    if value is None:
        return AbsentLastReloadTimestamp()
    val = isoparse(value)
    return PresentLastReloadTimestamp(val)



from datetime import time
from typing import Callable, List, TypeVar

from elimity_insights_client._util import encode_datetime, local_timezone
from elimity_insights_client.api.expression import (
    ActiveBooleanExpression,
    AggregateOperator,
    AllBooleanExpression,
    AnyBooleanExpression,
    AssignedBooleanExpression,
    AttributeBooleanExpression,
    AttributeDateExpression,
    AttributeDateTimeExpression,
    AttributeNumberExpression,
    AttributeStringExpression,
    AttributeTimeExpression,
    BooleanExpression,
    CmpOperator,
    DateCmpBooleanExpression,
    DateExpression,
    DateTimeCmpBooleanExpression,
    DateTimeExpression,
    DirectLinkAggregateNumberExpression,
    DirectlyLinkedToBooleanExpression,
    IdInBooleanExpression,
    IdStringExpression,
    LinkAggregateNumberExpression,
    LinkAssignedBooleanExpression,
    LinkAttributeBooleanExpression,
    LinkAttributeDateExpression,
    LinkAttributeDateTimeExpression,
    LinkAttributeNumberExpression,
    LinkAttributeStringExpression,
    LinkAttributeTimeExpression,
    LinkedToBooleanExpression,
    LiteralBooleanExpression,
    LiteralDateExpression,
    LiteralDateTimeExpression,
    LiteralNumberExpression,
    LiteralStringExpression,
    LiteralTimeExpression,
    MatchBooleanExpression,
    MatchMode,
    MatchOperator,
    NameStringExpression,
    NotBooleanExpression,
    NumberCmpBooleanExpression,
    NumberExpression,
    RelativeDateExpression,
    RelativeDateTimeExpression,
    StringExpression,
    TimeCmpBooleanExpression,
    TimeExpression,
)

_T = TypeVar("_T")
_EncodeFunc = Callable[[_T], object]


def encode_boolean_expression(expression: BooleanExpression) -> object:
    """Encode the given boolean expression to a JSON value."""
    if isinstance(expression, ActiveBooleanExpression):
        return {"reference": expression.reference, "type": "active"}

    if isinstance(expression, AllBooleanExpression):
        return _encode_composite_boolean_expression(expression.exprs, "all")

    if isinstance(expression, AnyBooleanExpression):
        return _encode_composite_boolean_expression(expression.exprs, "any")

    if isinstance(expression, AssignedBooleanExpression):
        return {
            "attributeType": expression.attribute_type,
            "reference": expression.reference,
            "type": "assigned",
        }

    if isinstance(expression, AttributeBooleanExpression):
        return _encode_attribute_expression(
            expression.attribute_type, expression.reference
        )

    if isinstance(expression, DateCmpBooleanExpression):
        return _encode_cmp_expression(
            encode_date_expression,
            expression.lhs,
            expression.operator,
            expression.rhs,
            "dateCmp",
        )

    if isinstance(expression, DateTimeCmpBooleanExpression):
        return _encode_cmp_expression(
            encode_date_time_expression,
            expression.lhs,
            expression.operator,
            expression.rhs,
            "dateTimeCmp",
        )

    if isinstance(expression, DirectlyLinkedToBooleanExpression):
        condition = encode_boolean_expression(expression.condition)
        return {
            "alias": expression.alias,
            "condition": condition,
            "entityType": expression.entity_type,
            "linkAlias": expression.link_alias,
            "sourceId": expression.source_id,
            "type": "directlyLinkedTo",
        }

    if isinstance(expression, IdInBooleanExpression):
        return {
            "ids": expression.ids,
            "reference": expression.reference,
            "type": "idInBooleanExpression",
        }

    if isinstance(expression, LinkedToBooleanExpression):
        condition = encode_boolean_expression(expression.condition)
        return {
            "alias": expression.alias,
            "condition": condition,
            "entityType": expression.entity_type,
            "sourceId": expression.source_id,
            "type": "linkedTo",
        }

    if isinstance(expression, LinkAssignedBooleanExpression):
        return {
            "linkAttributeType": expression.link_attribute_type,
            "linkReference": expression.link_reference,
            "type": "linkAssigned",
        }

    if isinstance(expression, LinkAttributeBooleanExpression):
        return _encode_link_attribute_expression(
            expression.link_attribute_type, expression.link_reference
        )

    if isinstance(expression, LiteralBooleanExpression):
        return {"boolean": expression.boolean, "type": "literal"}

    if isinstance(expression, MatchBooleanExpression):
        lhs = encode_string_expression(expression.lhs)
        mode = _encode_match_mode(expression.mode)
        operator = _encode_match_operator(expression.operator)
        rhs = encode_string_expression(expression.rhs)
        return {
            "lhs": lhs,
            "mode": mode,
            "operator": operator,
            "rhs": rhs,
            "type": "match",
        }

    if isinstance(expression, NotBooleanExpression):
        expr = encode_boolean_expression(expression.expr)
        return {"expr": expr, "type": "not"}

    if isinstance(expression, NumberCmpBooleanExpression):
        return _encode_cmp_expression(
            encode_number_expression,
            expression.lhs,
            expression.operator,
            expression.rhs,
            "numberCmp",
        )

    if isinstance(expression, TimeCmpBooleanExpression):
        return _encode_cmp_expression(
            encode_time_expression,
            expression.lhs,
            expression.operator,
            expression.rhs,
            "timeCmp",
        )


def encode_date_expression(expression: DateExpression) -> object:
    """Encode the given date expression to a JSON value."""
    if isinstance(expression, AttributeDateExpression):
        return _encode_attribute_expression(
            expression.attribute_type, expression.reference
        )

    if isinstance(expression, LinkAttributeDateExpression):
        return _encode_link_attribute_expression(
            expression.link_attribute_type, expression.link_reference
        )

    if isinstance(expression, LiteralDateExpression):
        date = expression.date.isoformat()
        return {"date": date, "type": "literal"}

    if isinstance(expression, RelativeDateExpression):
        return {
            "days": expression.days,
            "future": expression.future,
            "months": expression.months,
            "type": "relativeDate",
            "years": expression.years,
        }


def encode_date_time_expression(expression: DateTimeExpression) -> object:
    """Encode the given date-time expression to a JSON value."""
    if isinstance(expression, AttributeDateTimeExpression):
        return _encode_attribute_expression(
            expression.attribute_type, expression.reference
        )

    if isinstance(expression, LinkAttributeDateTimeExpression):
        return _encode_link_attribute_expression(
            expression.link_attribute_type, expression.link_reference
        )

    if isinstance(expression, LiteralDateTimeExpression):
        date_time = encode_datetime(expression.date_time)
        return {"dateTime": date_time, "type": "literal"}

    if isinstance(expression, RelativeDateTimeExpression):
        return {
            "days": expression.days,
            "future": expression.future,
            "hours": expression.hours,
            "minutes": expression.minutes,
            "months": expression.months,
            "seconds": expression.seconds,
            "type": "relativeDateTime",
            "years": expression.years,
        }


def encode_number_expression(expression: NumberExpression) -> object:
    """Encode the given number expression to a JSON value."""
    if isinstance(expression, AttributeNumberExpression):
        return _encode_attribute_expression(
            expression.attribute_type, expression.reference
        )

    if isinstance(expression, DirectLinkAggregateNumberExpression):
        condition = encode_boolean_expression(expression.condition)
        expr = encode_number_expression(expression.expr)
        op = _encode_aggregate_operator(expression.op)
        return {
            "alias": expression.alias,
            "condition": condition,
            "entityType": expression.entity_type,
            "linkAlias": expression.link_alias,
            "expr": expr,
            "op": op,
            "sourceId": expression.source_id,
            "type": "directLinkAggregate",
        }

    if isinstance(expression, LinkAggregateNumberExpression):
        condition = encode_boolean_expression(expression.condition)
        expr = encode_number_expression(expression.expr)
        op = _encode_aggregate_operator(expression.op)
        return {
            "alias": expression.alias,
            "condition": condition,
            "entityType": expression.entity_type,
            "expr": expr,
            "op": op,
            "sourceId": expression.source_id,
            "type": "linkAggregate",
        }

    if isinstance(expression, LinkAttributeNumberExpression):
        return _encode_link_attribute_expression(
            expression.link_attribute_type, expression.link_reference
        )

    if isinstance(expression, LiteralNumberExpression):
        return {"number": expression.number, "type": "literal"}


def encode_string_expression(expression: StringExpression) -> object:
    """Encode the given string expression to a JSON value."""
    if isinstance(expression, AttributeStringExpression):
        return _encode_attribute_expression(
            expression.attribute_type, expression.reference
        )

    if isinstance(expression, IdStringExpression):
        return {"reference": expression.reference, "type": "id"}

    if isinstance(expression, LinkAttributeStringExpression):
        return _encode_link_attribute_expression(
            expression.link_attribute_type, expression.link_reference
        )

    if isinstance(expression, LiteralStringExpression):
        return {"string": expression.string, "type": "literal"}

    if isinstance(expression, NameStringExpression):
        return {"reference": expression.reference, "type": "name"}


def encode_time_expression(expression: TimeExpression) -> object:
    """Encode the given time expression to a JSON value."""
    if isinstance(expression, AttributeTimeExpression):
        return _encode_attribute_expression(
            expression.attribute_type, expression.reference
        )

    if isinstance(expression, LinkAttributeTimeExpression):
        return _encode_link_attribute_expression(
            expression.link_attribute_type, expression.link_reference
        )

    if isinstance(expression, LiteralTimeExpression):
        time = _set_timezone(expression.time).isoformat()
        return {"time": time, "type": "literal"}


def _encode_aggregate_operator(operator: AggregateOperator) -> object:
    if operator is AggregateOperator.AVG:
        return "avg"

    if operator is AggregateOperator.COUNT:
        return "count"

    if operator is AggregateOperator.MAX:
        return "max"

    if operator is AggregateOperator.MIN:
        return "min"

    if operator is AggregateOperator.SUM:
        return "sum"


def _encode_attribute_expression(attribute_type: str, reference: str) -> object:
    return {
        "attributeType": attribute_type,
        "reference": reference,
        "type": "attribute",
    }


def _encode_cmp_operator(operator: CmpOperator) -> object:
    if operator is CmpOperator.EQ:
        return "eq"

    if operator is CmpOperator.GT:
        return "gt"

    if operator is CmpOperator.GTE:
        return "gte"

    if operator is CmpOperator.LT:
        return "lt"

    if operator is CmpOperator.LTE:
        return "lte"

    if operator is CmpOperator.NEQ:
        return "neq"


def _encode_composite_boolean_expression(
    exprs: List[BooleanExpression], type: str
) -> object:
    expr_iter = map(encode_boolean_expression, exprs)
    return {"exprs": expr_iter, "type": type}


def _encode_cmp_expression(
    func: _EncodeFunc[_T], lhs: _T, operator: CmpOperator, rhs: _T, type: str
) -> object:
    lhs_obj = func(lhs)
    operator_obj = _encode_cmp_operator(operator)
    rhs_obj = func(rhs)
    return {
        "lhs": lhs_obj,
        "operator": operator_obj,
        "rhs": rhs_obj,
        "type": type,
    }


def _encode_link_attribute_expression(
    link_attribute_type: str, link_reference: str
) -> object:
    return {
        "linkAttributeType": link_attribute_type,
        "linkReference": link_reference,
        "type": "linkAttribute",
    }


def _encode_match_mode(mode: MatchMode) -> object:
    if mode is MatchMode.CASE_INSENSITIVE:
        return "caseInsensitive"

    if mode is MatchMode.CASE_SENSITIVE:
        return "caseSensitive"


def _encode_match_operator(operator: MatchOperator) -> object:
    if operator is MatchOperator.CONTAINS:
        return "contains"

    if operator is MatchOperator.ENDS_WITH:
        return "endsWith"

    if operator is MatchOperator.EQUALS:
        return "equals"

    if operator is MatchOperator.STARTS_WITH:
        return "startsWith"


def _set_timezone(time: time) -> time:
    if time.tzinfo is None:
        return time.replace(tzinfo=local_timezone)
    return time



from typing import List

from elimity_insights_client.api._api import Config
from elimity_insights_client.api._api import query as api_query
from elimity_insights_client.api._api import sources
from elimity_insights_client.api.entities._entity import Entity
from elimity_insights_client.api.entities._parse_query_results_page import (
    parse_query_results_page,
)
from elimity_insights_client.api.entities._query import query


def entities(config: Config, entity_type: str, source_id: int) -> List[Entity]:
    """
    List all entities of the given entity type from the given source.

    The resulting entities also include all attribute assignments, and links for every other entity type of the given
    source.
    """

    sos = sources(config)
    schema = next(
        source.domain_graph_schema for source in sos if source.id == source_id
    )
    que = query(entity_type, schema, source_id)
    queries = [que]
    (page,) = api_query(config, queries)
    return parse_query_results_page(entity_type, page, schema)



from dataclasses import dataclass
from typing import List

from more_itertools import chunked

from elimity_insights_client._domain_graph_schema import (
    AttributeType,
    DomainGraphSchema,
)
from elimity_insights_client._util import map_list
from elimity_insights_client.api.entities._entity import Entity, Link
from elimity_insights_client.api.entities._schema import (
    attribute_types as schema_attribute_types,
)
from elimity_insights_client.api.entities._schema import (
    link_entity_types as schema_link_entity_types,
)
from elimity_insights_client.api.query_results_page import (
    BooleanValue,
    QueryResult,
    QueryResultsPage,
    Value,
)


@dataclass
class _AttributeAssignment:
    assigned: bool
    attribute_type: str
    value: Value


@dataclass
class _LinkGroup:
    entity_type: str
    links: List[Link]


def parse_query_results_page(
    entity_type: str, page: QueryResultsPage, schema: DomainGraphSchema
) -> List[Entity]:
    def make_entity(result: QueryResult) -> Entity:
        entity = _link(entity_type, schema, result)

        def make_group(entity_type: str, page: QueryResultsPage) -> _LinkGroup:
            def make_link(result: QueryResult) -> Link:
                return _link(entity_type, schema, result)

            links = map_list(make_link, page.results)
            return _LinkGroup(entity_type, links)

        link_entity_types = schema_link_entity_types(entity_type, schema)
        group_iter = map(make_group, link_entity_types, result.link_pages)
        links = {group.entity_type: group.links for group in group_iter}
        return Entity(entity.attribute_assignments, entity.id, links, entity.name)

    return map_list(make_entity, page.results)


def _attribute_assignment(
    inclusions: List[Value], type: AttributeType
) -> _AttributeAssignment:
    assigned_inclusion, value_inclusion = inclusions
    assigned = assigned_inclusion == BooleanValue(True)
    return _AttributeAssignment(assigned, type.id, value_inclusion)


def _link(entity_type: str, schema: DomainGraphSchema, result: QueryResult) -> Link:
    inclusions_iter = chunked(result.inclusions, 2)
    attribute_types = schema_attribute_types(entity_type, schema)
    assignment_iter = map(_attribute_assignment, inclusions_iter, attribute_types)
    assignments = {
        assignment.attribute_type: assignment.value
        for assignment in assignment_iter
        if assignment.assigned
    }
    entity = result.entity
    return Link(assignments, entity.id, entity.name)



"""
Client for connector interactions with an Elimity Insights server.

Note that this module interprets timestamps without timezone information as
being defined in the local system timezone.
"""

from elimity_insights_client._domain_graph_schema import (
    AttributeType,
    DomainGraphSchema,
    EntityType,
    RelationshipAttributeType,
    Type,
)
from elimity_insights_client._elimity_insights_client import (
    AttributeAssignment,
    BooleanValue,
    Certificate,
    Client,
    Config,
    ConnectorLog,
    DateTime,
    DateTimeValue,
    DateValue,
    DomainGraph,
    Entity,
    Level,
    NumberValue,
    Relationship,
    StringValue,
    TimeValue,
    Value,
)

__all__ = [
    "AttributeAssignment",
    "AttributeType",
    "BooleanValue",
    "Certificate",
    "Client",
    "Config",
    "ConnectorLog",
    "DateTime",
    "DateTimeValue",
    "DateValue",
    "DomainGraph",
    "DomainGraphSchema",
    "Entity",
    "EntityType",
    "Level",
    "NumberValue",
    "Relationship",
    "RelationshipAttributeType",
    "StringValue",
    "TimeValue",
    "Type",
    "Value",
]



from typing import List

from typing_extensions import NotRequired, TypedDict

from elimity_insights_client._domain_graph_schema import (
    AttributeType,
    DomainGraphSchema,
    EntityType,
    RelationshipAttributeType,
    Type,
)
from elimity_insights_client._util import map_list

AttributeTypeDict = TypedDict(
    "AttributeTypeDict",
    {
        "archived": bool,
        "description": str,
        "entityTypeId": str,
        "id": str,
        "name": str,
        "type": str,
    },
)
EntityTypeDict = TypedDict(
    "EntityTypeDict",
    {
        "anonymized": bool,
        "icon": str,
        "id": str,
        "plural": str,
        "singular": str,
    },
)
RelationshipAttributeTypeDict = TypedDict(
    "RelationshipAttributeTypeDict",
    {
        "archived": bool,
        "childType": str,
        "description": NotRequired[str],
        "id": str,
        "name": str,
        "parentType": str,
        "type": str,
    },
)
DomainGraphSchemaDict = TypedDict(
    "DomainGraphSchemaDict",
    {
        "entityAttributeTypes": List[AttributeTypeDict],
        "entityTypes": List[EntityTypeDict],
        "relationshipAttributeTypes": List[RelationshipAttributeTypeDict],
    },
)


def decode_domain_graph_schema(json: DomainGraphSchemaDict) -> DomainGraphSchema:
    """Decode the given JSON value to a domain graph schema."""
    attribute_types = json["entityAttributeTypes"]
    attribute_types_ = map_list(_decode_attribute_type, attribute_types)
    entity_types = json["entityTypes"]
    entity_types_ = map_list(_decode_entity_type, entity_types)
    relationship_attribute_types = json["relationshipAttributeTypes"]
    relationship_attribute_types_ = map_list(
        _decode_relationship_attribute_types, relationship_attribute_types
    )
    return DomainGraphSchema(
        attribute_types_, entity_types_, relationship_attribute_types_
    )


def _decode_attribute_type(json: AttributeTypeDict) -> AttributeType:
    archived = json["archived"]
    description = json["description"]
    entity_type = json["entityTypeId"]
    id_ = json["id"]
    name = json["name"]
    type_ = json["type"]
    type__ = _decode_type(type_)
    return AttributeType(archived, description, entity_type, id_, name, type__)


def _decode_entity_type(json: EntityTypeDict) -> EntityType:
    anonymized = json["anonymized"]
    icon = json["icon"]
    id = json["id"]
    plural = json["plural"]
    singular = json["singular"]
    return EntityType(anonymized, icon, id, plural, singular)


def _decode_relationship_attribute_types(
    json: RelationshipAttributeTypeDict,
) -> RelationshipAttributeType:
    archived = json["archived"]
    description = json.get("description", "")
    from_entity_type = json["parentType"]
    id_ = json["id"]
    name = json["name"]
    to_entity_type = json["childType"]
    type_ = json["type"]
    type__ = _decode_type(type_)
    return RelationshipAttributeType(
        archived, description, from_entity_type, id_, name, to_entity_type, type__
    )


def _decode_type(json: str) -> Type:
    if json == "boolean":
        return Type.BOOLEAN
    elif json == "date":
        return Type.DATE
    elif json == "dateTime":
        return Type.DATE_TIME
    elif json == "number":
        return Type.NUMBER
    elif json == "string":
        return Type.STRING
    else:
        return Type.TIME



"""Endpoints for API interactions with an Elimity Insights server."""

from dataclasses import dataclass
from typing import List, Optional, Type, TypeVar, cast

from requests import request

from elimity_insights_client._util import encoder, map_list
from elimity_insights_client.api._decode_query_results_page import (
    QueryResultsPageDict,
    decode_query_results_page,
)
from elimity_insights_client.api._decode_source import SourceDict, decode_source
from elimity_insights_client.api._encode_query import encode_query
from elimity_insights_client.api.query import Query
from elimity_insights_client.api.query_results_page import QueryResultsPage
from elimity_insights_client.api.source import Source


@dataclass
class Config:
    """Configuration consisting of API credentials and connection properties."""

    token_id: str
    token_secret: str
    url: str
    verify_ssl: bool


_T = TypeVar("_T")


def query(config: Config, queries: List[Query]) -> List[QueryResultsPage]:
    """Perform the given queries and return the result pages."""
    query_iter = map(encode_query, queries)
    data = encoder.encode(query_iter)
    page_dicts = _request(
        config, data, "POST", "/api/agent/query", List[QueryResultsPageDict]
    )
    return map_list(decode_query_results_page, page_dicts)


def sources(config: Config) -> List[Source]:
    """List all configured sources."""
    source_dicts = _request(config, None, "GET", "/api/agent/sources", List[SourceDict])
    return map_list(decode_source, source_dicts)


def _request(
    config: Config, data: Optional[str], method: str, path: str, _type: Type[_T]
) -> _T:
    auth = config.token_id, config.token_secret
    headers = {"Content-Type": "application/json"}
    response = request(
        method,
        config.url + path,
        auth=auth,
        data=data,
        headers=headers,
        verify=config.verify_ssl,
    )
    response.raise_for_status()
    json = response.json()
    return cast(_T, json)



"""Type definitions for query result pages."""

from dataclasses import dataclass
from datetime import date, datetime, time
from typing import List, Union


@dataclass
class BooleanValue:
    """Typed wrapper for a boolean value."""

    value: bool


@dataclass
class DateValue:
    """Typed wrapper for a date value."""

    value: date


@dataclass
class DateTimeValue:
    """Typed wrapper for a date-time value."""

    value: datetime


@dataclass
class Entity:
    """Matched entity as part of a query result."""

    active: bool
    id: str
    name: str


@dataclass
class GroupByQueryResult:
    """Matched group as part of a group-by query result page."""

    count: int
    label: "Value"
    sub_pages: List["GroupByQueryResultsPage"]


@dataclass
class GroupByQueryResultsPage:
    """Page of results for a group-by query."""

    group_count: int
    results: List[GroupByQueryResult]


@dataclass
class NumberValue:
    """Typed wrapper for a number value."""

    value: float


@dataclass
class QueryResult:
    """Entry in a page of query results consisting of entity details, inclusions and sub-results."""

    entity: Entity
    inclusions: List["Value"]
    link_group_by_pages: List[GroupByQueryResultsPage]
    link_pages: List["QueryResultsPage"]


@dataclass
class QueryResultsPage:
    """Page of results for a query."""

    count: int
    results: List[QueryResult]


@dataclass
class StringValue:
    """Typed wrapper for a string value."""

    value: str


@dataclass
class TimeValue:
    """Typed wrapper for a time value."""

    value: time


Value = Union[
    BooleanValue, DateValue, DateTimeValue, NumberValue, StringValue, TimeValue
]



from dataclasses import dataclass
from typing import Dict, List

from elimity_insights_client.api.query_results_page import Value


@dataclass
class Entity:
    """Single entity with assignments for all attributes and links for all entity types."""

    attribute_assignments: Dict[str, Value]
    id: str
    links: Dict[str, List["Link"]]
    name: str


@dataclass
class Link:
    """Linked entity with assignments for all attributes."""

    attribute_assignments: Dict[str, Value]
    id: str
    name: str
